// Processing 3.4 - http://processing.org/

final int ysplit = 207, birdSize = 0xa00, bg=16, hudoff=0x30,
          magx = 100, magy = 74, magscale = 3, RED=#ff0000, GREEN=#00ff00;
final int magw = magx<<magscale, magh = magy<<magscale;
final int[][] moves = {{-0x180, -0x80, 0, 0x80, 0x180}, // [0] moves considered valid
             {#ff00af, #cf007f, -1, #7f00ff, #776699}}; // [1] and their colors

int birdxlo;      // 0 inherited 'xsub', configurable: (0..15)*16
int ryuPos;       // 1 rw(0x85)
int screen;       // 2 rws(0x51)*0x100 + rb(0x50)
int origin;       // offset for where to start drawing
int roomSize;     // 3 rb(0x53)*0x10000
int goleft;       // 4 offset load seams for right-to-left scrolling rooms ($3d.7)
int screenP;      // 5 previous value of (screen-goleft)&0xf000 to detect load seams
int birdPos;      // 6 rw(0x458+bs,0x460+bs)
int birdY;        // 7 rb(0x480+bs)
int birdVel;      // 8 rws(0x448+bs,0x450+bs) --or just 0 if starting from bird spawn frame
boolean loading;  // 9 true every 16 pixels scrolled (0x1000's digit of screen changes)
boolean scroll;   // 10 true until roomSize exceeded (left ends of rooms not implemented)
boolean collision;// true if bird, ninja are in collision range horizontally
boolean isbird;   // false when bird is scrolled off screen
int ryuVel;       // taken from ninja[].z each loop, acceptable values from moves[]
int ninjY, ninjX, birdX, k_x, k_y, k_y2, m_x, m_y, imgwidth, frame, f, target;
int[] tiles;      // [0] = screen pos @ end; odd = x; even = y
String[] input, spd, yps, bgc;
PVector[] ninja, bird;  // x = x, y = y, z = ... x speed
PImage original, current, previous;
boolean compare = false;
boolean update = true;
boolean init = true;
boolean magtog = false;
String colstr = "";

void settings() {
  size(1600, 800);
  noSmooth();
}

void setup() {
  frameRate(60);
  noFill();
  textSize(16);
  textAlign(RIGHT);
  input = loadStrings("birdOut.txt");
  LoadSettings();
}

void draw() { // main
  background(bg);
  if (update) Redraw();
  image(compare ? previous : current, 0, 0);
  DrawUI();
}

void DrawUI() {
  m_x = mouseX-(magx>>1);
  m_y = mouseY-(magy>>1);
  k_x = int(ninja[f].x)-(magx>>1);
  k_y = int(ninja[f].y)-(magy>>1);
  k_y2 = ysplit-15-(f&15);

  stroke(#66ff66, 20);
  line(ninja[f].x,0, ninja[f].x,ysplit-1); // vertical line highlight 
  stroke(#66ff66, 160);
  rect(ninja[f].x-1, ninja[f].y-1, 2, 2);  // upper pixel highlight
  rect(ninja[f].x-1, k_y2-1, 2, 2);        // lower pixel highlight
  rect(mouseX-1, mouseY-1, 2, 2);          // mouse pointer hightlight

  copy(k_x, magtog?k_y:ysplit-magy, magx, magy, 0, ysplit+20, magw, magh);  // L: magnify selected region

  stroke(160, 184);
  line(0, ysplit, width, ysplit);
  line(magw, ysplit, magw, height);

  if (m_y < ysplit) {
    rect(m_x-1, m_y-1, magx+1, magy+1);  // pointer mag. region outline
    copy(m_x, m_y, magx, magy, magw+1, ysplit+1, magw, magh); // R: magnify m-pointer region
  }

  rect(k_x-1, magtog?k_y-1:ysplit-magy, magx+1, magy+1); // active mag. region outline 

  int ys = ysplit+17, tx=130;
  text(compare ? f+" C" : f+"", magw-4, ys);
  text("frame "+(f+frame), tx, ys);
  text("dist. "+(int(ninja[f].x) - int(bird[f].x)), tx+=100, ys);
  text("Bv "+String.format("%.2f",bird[f].z/256), tx+=100, ys);
  text("Rv "+ninja[f].z/256, tx+=100, ys);
  textSize(10);
  text(colstr, width-10, 20);
  textSize(16);
}

void Redraw() {
  //ResetVars();
  DrawBG();
  for (int i = 0; i < ninja.length; i++) {

    int scrn = (screen-origin);
    birdX = (birdPos+scrn)>>8;
    ninjX = (ryuPos+scrn)>>8;
    ninjY = int(ninja[i].y);
    ninja[i].x = ninjX;
    bird[i].x = birdX;
    bird[i].y = birdY;
    bird[i].z = birdVel;
    boolean ahead = (birdX >= ninjX);
    if (loading) Lmark(ninjX);

    if (collision) {
      int c = ahead ? RED : GREEN;
      stroke(c, 60);  // green = boosted rightward
      line(ninjX, ninjY, birdX, birdY); // collision vis., connecting line
      stroke(c); // won't show up with any level of transparency what's up with that
      point(ninjX, ysplit-32); // collision vis., lower
      if (!ahead && i >= target) 
        {colstr = colstr+(i+frame)+", "+((ryuPos+screen)>>8)+"\n";}
    }

    stroke(ahead ? GREEN : RED); // ahead = green, bird behind ryu = red
    point(birdX, birdY);        // draw bird path

    ryuVel = int(ninja[i].z);
    stroke(GetColor());
    point(ninjX, ninjY);            // draw ninja path, upper
    point(ninjX, ysplit-15-(i&15)); // draw ninja path, lower

    GameLoop();
  }

  text(birdxlo/16, 24, 17);
  loadPixels();  // keep a pixel buffer and only update it when necessary
  for (int y = 0; y < current.height; y++) { // and don't copy the whole width
    for (int x = 0; x < current.width; x++) {
      current.pixels[y*current.width+x] = get(x,y);
    }
  }
  current.updatePixels();
  if (init) FirstCopy();
  update = false;
}

void GameLoop() {

  if (scroll) {
    screen = screen + ryuVel;
    birdPos = birdPos - ryuVel;
  } else ryuPos = ryuPos + ryuVel;

  int scrtmp = (screen-goleft)&0xf000;
  loading = (scrtmp != screenP);
  screenP = scrtmp;
  scroll = (screen < roomSize) && (screen >= 0);

  if (!loading && isbird) {
    if (birdPos >= ryuPos) {
      birdVel = birdVel - 0x10;
    } else {
      birdVel = birdVel + 0x10;
    }

    birdPos = birdPos + birdVel;
    isbird = (birdPos <= 0xf800) && (birdPos >= 0x0800);

    if (birdY >= ninjY) {
      birdY--;
    } else {
      birdY++;
    }

    collision = Bump();
  }
}

boolean Bump() {  // to mark where bird+ninja are in x collision range
  if (ryuPos >= birdPos) {  // is this even accurate? yeah, seems OK.
    if (ryuPos-birdPos <= birdSize) return true;
  } else {
    if (birdPos-ryuPos <= birdSize) return true;
  }
  return false;
}

void DrawBG() {
  stroke(255, 42);
  int s = (tiles[0]-origin)>>8;
  for (int i = 1; i < tiles.length; i+=2)
    rect(tiles[i]+s, tiles[i+1]+1-hudoff, 14, 14);
}

void Lmark(int x) {  // mark background loading seams every 16 pixels
  stroke(108, 80);
  line(x, 0, x, ysplit-1);
}

int GetColor() { // color code ninja pixels by movement
  for (int i = 0; i < moves[0].length; i++) {
    if (moves[0][i] == ryuVel) {
      int c = (goleft == 0) ? i : (-i+4)%5; // reverse order for R-to-L scroller
      return moves[1][c];
    }
  }
  return 160;  // anything not in the list is grey
}
